;**********************************************************************
;                                                                     *
;    Filename:	    irtx.asm                                          *
;    Date:                                                            *
;    File Version:                                                    *
;                                                                     *
;    Author:        el@jap.hu                                         *
;    Page:          http://jap.hu/electronic/                         *
;**********************************************************************
;NOTES
;
; infrared transmitter modulator
;
; commands:
;
; SETSTATE: set IR mod output state
; Setting the mod output HIGH for a long time can result in
; blowing the IR transmitter LED out!
; 54 <state>
;
; SETPARAMS: set coding parameters
; 55 ir_T head_h head_l bit0_h bit0_l bit1_h bit1_l tail_h tail_l
; 55 ir_T skipbits togglebits firstbyte rc5tail
;
; SETPWM: set pwm parameters
; 58 pwm_T dc_T
; pwm T =pwm_T * 1 usec; duty cycle = dc_T * 0.25 usec
;
; SENDPDM: send ir command
;  header pulse high for T*head_h, low for T*head_l
;  data bytes transmitted as: bit0 bit1 bit2 bit3 bit4 bit5 bit6 bit7
;  tail bit is high for T*tail_h
;  bit0 is high for T*bit0_h, low for T*bit0_l
;  bit1 is high for T*bit1_h, low for T*bit1_l
; 56 <byte> [byte] [byte] ...
;
; SENDRC5: send ir command
;  transmit 2 bytes per ir command: <firstbyte> 8-skipbits bits
;  then <secondbyte>, all 8 bits
;  data bytes are transmitted as: bit7 bit6 bit5 bit4 bit3 bit2 bit1 bit0
;  bit0 is high for T, low for T
;  bit1 is low for T, high for T
;
;  This command can be used to repeat a given ir command, but not to
;  transmit more commands one by one, because the toggle bit is
;  only set after sending out all the ir commands.
;  To send more than one ir commands, repeat the SENDRC5 command with
;  the appropriate second byte(s).
; 57 <second byte> [second byte] ...
;
;
; ir_T is given in 10 usecs
; other timing values are given in ir_T steps
;
;**********************************************************************
;HISTORY
; 001 - 20020615 inital
; 002 - 20020615
; 003 - 20020615 working PDM sender
; 004 - 20020616
; 005 - 20120902 add PWM modulation on output
; 006 - 20121014 add PWM setup command
;
;**********************************************************************

	list      p=16f627a

	__CONFIG   _CP_OFF & _WDT_OFF & _PWRTE_ON & _HS_OSC & _LVP_OFF & _MCLRE_OFF

#include <p16f627a.inc>
#include "serial2.inc"

;***** CONSTANT VALUES
buffer1size	EQU	.30
buffer2size	EQU	.10

;*** PWM output
pwm_freq	EQU d'36000'
dc_value	EQU (d'1000000' / pwm_freq) >> 1
dc1b_value	EQU ( (d'1000000' / pwm_freq) << 1) - (dc_value << 2)

ccp1con_value	EQU (1<<CCP1M3) | (1<<CCP1M2) | (dc1b_value << 4)
pr2_value	EQU (d'1000000' / pwm_freq) - 1


;***** VARIABLES

.mainda UDATA
savew1	res 1
savestatus res 1
savepclath res 1
savefsr res 1

;readline
bufptr	res 1
bufpt2	res 1
serchr	res 1
rdhex	res 1
bytecnt	res 1
buffer1	res .30
#define buffer2 buffer1

; ir specific data
ir_head_h	res 1
ir_head_l	res 1
ir_bit0_h	res 1
ir_bit0_l	res 1
ir_bit1_h	res 1
ir_bit1_l	res 1
ir_tail_h	res 1
ir_tail_l	res 1
ir_T		res 1 ; T in 10 usec steps

#define ir_skipbits	ir_head_h
#define ir_togglebits	ir_head_l
#define ir_firstbyte	ir_bit0_h
#define ir_rc5tail	ir_bit0_l

ir_cnt1		res 1
ir_cnt2		res 1
ir_byte		res 1
ir_bitcnt	res 1

pr2_reg		res 1
ccp1con_reg	res 1

vectors		CODE 0
  		goto    main
		nop
		nop
		nop

		goto itr

prog		CODE 5

sbus_cmd_vectors
		addlw sbus_cmd
		movwf PCL
sbus_cmd	; command jump table

		goto cmd_setstate ; 0x54: set output port state
		goto cmd_setparam ; 0x55: set parameters
		goto cmd_sendpdm  ; 0x56: send pdm modulated data
		goto cmd_sendrc5  ; 0x57: send rc5 encoded data
		goto cmd_setpwm   ; 0x58: set pwm parameters

itr
	movwf	savew1
	movf	STATUS,w
	clrf	STATUS
	movwf	savestatus
	movf	PCLATH,w
	movwf	savepclath
	clrf	PCLATH
	movf	FSR,w
	movwf	savefsr

	call uart_int_handler

	movf	savefsr,w
	movwf	FSR
	movf	savepclath,w
	movwf	PCLATH
	movf	savestatus,w
	movwf	STATUS
	swapf	savew1,f
	swapf	savew1,w
	retfie

main

	bcf	STATUS,RP0
	movlw 7
	movwf CMCON		; no comparators, just digital I/O

gie01	bcf	INTCON,GIE	;turn GIE off 
	btfsc	INTCON,GIE
	goto 	gie01

	clrf	PIR1		;clear peripheral flags
 
	clrf	PORTA		;clear all i/o registers...
	clrf	PORTB

 	bsf     STATUS,RP0	;allow access to page 1 stuff!
	clrf	PIE1		;disable peripheral interrupts

	movlw 0x2 ; RB1 = RX input, RB2=TX output
	movwf TRISB
	clrf	INTCON
	clrf	PIE1
	bcf	STATUS, RP0

	call ir_init
	call uart_init

warm
	movlw 1
	movwf ir_T
	movwf ir_head_h
	movwf ir_head_l
	movwf ir_bit0_h
	movwf ir_bit0_l
	movwf ir_bit1_h
	movwf ir_bit1_l
	movwf ir_tail_h
	movwf ir_tail_l

loop
	movlw 0x0d ; send DOS-style CRLF line end
	call uart_tx
	movlw 0x0a
	call uart_tx
	movlw 'R'
	call uart_tx
	movlw 'D'
	call uart_tx
	movlw ' '
	call uart_tx

	call readline ; read line into buffer1, length=bufptr

	clrf bufpt2
	clrf bytecnt ; bytes read
loop3	clrf rdhex
loop2	movf bufpt2, W
	xorwf bufptr, W ; end of line?
	bz loop_eol

	movlw buffer1
	addwf bufpt2, W
	movwf FSR
	incf bufpt2, F
	movf INDF, W
	xorlw 0x20
	bz loop_spc
	xorlw 0x20
	movwf serchr
	sublw 0x40
	movlw 0x09
	btfss STATUS, C ; skip if number
	addwf serchr, F
loop_hx2 movlw 0x0f
	andwf serchr, F
	swapf rdhex, W
	andlw 0xf0
	iorwf serchr, W
	movwf rdhex
	goto loop2

loop_spc ; go to next hex byte

	movf serchr, W
	xorlw 0x20
	bz loop2

	call write_char
	goto loop3

loop_eol ; end of line

	call write_char
	; bytecnt bytes read into buffer2

	movf bytecnt, W
	xorlw 0
	bz loop ; no bytes read

	movlw 1
	movwf bufpt2 ; actual byte to read

	call sbus_cmd_handler
	goto loop

sbus_cmd_handler
	; command is in buffer2
	movlw 0x54
	subwf buffer2, F ; buffer2 -= 0x54
	bnc cext ; result is negative

	movlw 0x05
	subwf buffer2, W ; if (buffer2 >=5) goto cext
	bc cext

	movf buffer2, W
	goto sbus_cmd_vectors

cext	movlw 'N'
	call uart_tx
	movlw 'I'
	call uart_tx
	return

readbyte movf bytecnt,W ; return Z=1 if no byte is read
	subwf bufpt2, W
	bz rb0 ; buffer2 read
	movlw buffer2
	addwf bufpt2, W
	movwf FSR
	movf INDF, W
	incf bufpt2, F
	bcf STATUS, Z
rb0	return

write_char
	movlw buffer2size
	subwf bytecnt, W
	bz bfull

	movlw buffer2
	addwf bytecnt, W
	movwf FSR
	movf rdhex, W
	movwf INDF
	incf bytecnt, F
bfull	return

readline
	clrf bufptr
rd0	call uart_rx
	btfsc STATUS, Z
	goto rd0
	movwf serchr
	xorlw 0x08 ; (backspace)
	btfsc STATUS, Z	
	goto rddel
	movf serchr, W
	xorlw 0x0d ; (enter)
	btfsc STATUS, Z
	goto rdent
	; other char just store
	movlw buffer1size
	subwf bufptr, W
	bz rd0 ; buffer full, dont store

	movlw buffer1
	addwf bufptr, W
	movwf FSR
	movf serchr, W
	movwf INDF
	call uart_tx

	incf bufptr, F
	goto rd0

rddel	movf bufptr, W
	xorlw 0
	bz rd0 ; no char to del
	decf bufptr, F
	movf serchr, W
	call uart_tx
	movlw ' '
	call uart_tx
	movf serchr, W
	call uart_tx
	goto rd0

rdent	movf serchr, W
	call uart_tx
	movlw 0x0a
	call uart_tx

	movf bufptr, W ; enter, return
	return


; ir specific code is here

cmd_miss ; required parameter missing
	movlw 'E'
	call uart_tx
	movlw 'R'
	goto uart_tx

cmd_setstate ; set output port state
	call readbyte
	bz cmd_miss
	andlw 0xff
	bz clear_mod

set_mod	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

printok movlw 'O'
	call uart_tx
	movlw 'K'
	goto uart_tx

clear_mod
	clrf CCP1CON
	goto printok

cmd_setparam ; set parameters
	call readbyte
	bz cmd_miss
	movwf ir_T

	call readbyte
	bz cmd_set0
	movwf ir_head_h

	call readbyte
	bz cmd_set0
	movwf ir_head_l

	call readbyte
	bz cmd_set0
	movwf ir_bit0_h

	call readbyte
	bz cmd_set0
	movwf ir_bit0_l

	call readbyte
	bz cmd_set0
	movwf ir_bit1_h

	call readbyte
	bz cmd_set0
	movwf ir_bit1_l

	call readbyte
	bz cmd_set0
	movwf ir_tail_h

	call readbyte
	bz cmd_set0
	movwf ir_tail_l

cmd_set0 goto printok

ir_init	; initialize IR module
	clrf CCP1CON

	; PWM init
	BANKSEL T2CON
	movlw (1<<TMR2ON) | (0x00)
	movwf T2CON ; TMR2 on, 1:1 prescaler
	movlw dc_value
	movwf CCPR1L ; PWM duty cycle
	
	clrf CCP1CON ; turn off CCP module

	BANKSEL PR2
	movlw pr2_value
	movwf PR2
	BANKSEL PORTA
	movlw ccp1con_value
	movwf ccp1con_reg
	movlw 3
	sublw pr2_value
	movwf pr2_reg

	return

cmd_setpwm
	call readbyte
	bz cmd_miss
	andlw 0xff ; set Z
	btfss STATUS, Z
	addlw 0xff
	BANKSEL PR2
	movwf PR2 ; PR2 = pwm_T - 1 (PWM period)
	BANKSEL PORTA
	addlw 0xfd
	movwf pr2_reg ; pr2_reg = PR2 - 3

	call readbyte
	bz cmd_pwmdone
	movwf ir_cnt1
	rrf ir_cnt1, F
	rrf ir_cnt1, W
	andlw 0x3f
	movwf CCPR1L ; PWM duty cycle (higher bits)
	swapf INDF, W
	andlw 0x30 ; preserve bits 5-4
	movwf ir_cnt1
	iorlw (1<<CCP1M3) | (1<<CCP1M2)
	movwf ccp1con_reg
	movf CCP1CON, F ; PWM is on?
	btfss STATUS, Z
	movwf CCP1CON ; update PWM duty cycle (lower bits)

cmd_pwmdone goto printok
	

cmd_sendpdm  ; send pdm modulated data
ir_sendpdm ; send PDM data

	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

	movf ir_head_h, W
	call ir_delay
	clrf CCP1CON

	movf ir_head_l, W
	call ir_delay

	; send out data
ir_send0
	call readbyte
	bz ir_send1

	movwf ir_byte
	movlw 8
	movwf ir_bitcnt

ir_send2 ; send out bits, one by one

	rrf ir_byte, F
	btfss STATUS, C
	goto ir_send3

ir_pdmbit1
	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

	movf ir_bit1_h, W
	call ir_delay
	clrf CCP1CON
	movf ir_bit1_l, W
	call ir_delay
	goto ir_send4

ir_send3
ir_pdmbit0
	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

	movf ir_bit0_h, W
	call ir_delay
	clrf CCP1CON

	movf ir_bit0_l, W
	call ir_delay

ir_send4
	decfsz ir_bitcnt, F
	goto ir_send2

	goto ir_send0


ir_send1 ; send last pulse

	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

	movf ir_tail_h, W
	call ir_delay
	clrf CCP1CON
	goto printok

cmd_sendrc5  ; send rc5 encoded data
ir_sendrc5 ; send RC5 data

	movf ir_firstbyte, W
	movwf ir_byte
	movlw 8
	movwf ir_bitcnt
	movf ir_skipbits, W
	movwf ir_cnt1

	; skip extra bits
ir_send5
	rlf ir_byte, F
	decf ir_bitcnt, F
	decfsz ir_cnt1, F
	goto ir_send5

	; transmit first byte
	call ir_rc5tx

	call readbyte
	bz ir_send6 ; done

	; transmit second byte
	movwf ir_byte
	movlw 8
	movwf ir_bitcnt
	call ir_rc5tx

	clrf CCP1CON

	call readbyte
	bz ir_send6 ; no more bytes, done
	decf bufpt2, F

	; sleep before next transmission
	movf ir_rc5tail, W
	call ir_delay
	goto ir_sendrc5

ir_send6
	clrf CCP1CON
	; invert the toggle bit in first byte
	movf ir_togglebits, W
	xorwf ir_firstbyte, F
	goto printok

ir_rc5tx
	rlf ir_byte, F
	btfss STATUS, C
	goto ir_rc5bit0

ir_rc5bit1
	clrf CCP1CON
	movlw 1
	call ir_delay
	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

	movlw 1
	call ir_delay
	goto ir_rc5tx2

ir_rc5bit0
	movf pr2_reg, W
	movwf TMR2
	movf ccp1con_reg, W
	movwf CCP1CON

	movlw 1
	call ir_delay
	clrf CCP1CON

	movlw 1
	call ir_delay

ir_rc5tx2
	decfsz ir_bitcnt, F
	goto ir_rc5tx
	return

ir_delay	; delay W * T * 10 usec
	movwf ir_cnt2

ir_delay_t	; delay t * 10 usec
	movf ir_T, W
	movwf ir_cnt1
ir_delay0 nop
	nop
	nop
	nop
	nop
	nop
	nop
	decfsz ir_cnt1, F
	goto ir_delay0

	decfsz ir_cnt2, F
	goto ir_delay_t

	return


	end

